#include "MotionControl.h"
#include <VirtualRobot/IK/GenericIKSolver.h>
#include <VirtualRobot/IK/DifferentialIK.h>
#include <VirtualRobot/Nodes/RobotNode.h>
#include <VirtualRobot/MathTools.h>
#include <Core/robotstate/remote/RemoteRobot.h>
#include <Core/observers/variant/ChannelRef.h>
#include <Core/observers/ConditionCheck.h>

#include <Core/core/system/ArmarXDataPath.h>

#include <Eigen/src/Geometry/Quaternion.h>

using namespace armarx;
using namespace armarx::MoveHelpers;

MotionControl::MotionControl()
{
}



void MotionControl::onInitRemoteStateHandler()
{
    addState<MoveJoints>("MoveJoints");
    addState<MotionControlTestState>("MotionControlTestState");
}



void MotionControl::onConnectRemoteStateHandler()
{
}



MoveJoints::MoveJoints()
{
}



void MoveJoints::defineParameters()
{
    context = getContext<RobotStatechartContext>();

    std::string configFilepathAbsolute;
    ArmarXDataPath::getAbsolutePath("HumanoidRobotAPIConfigs/stateconfigs/MotionControl.xml", configFilepathAbsolute);
    setConfigFile(configFilepathAbsolute);

    addToInputAsList("jointNames", VariantType::String, false);
    addToInputAsList("targetJointValues", VariantType::Float, false);

    // TODO: add when we have decided how to do it
    addToInputAsList("jointMaxSpeeds", VariantType::Float, true);
    addToInput("jointMaxSpeed", VariantType::Float, true);

    addToInput("targetTolerance", VariantType::Float, false);

    addToInput("timeoutInMs", VariantType::Int, false);

    addToOutputAsList("jointDistancesToTargets", VariantType::Float, true);
}



void MoveJoints::onEnter()
{
    NameValueMap targetJointAngles;
    SingleTypeVariantListPtr jointNames = getInputList("jointNames");
    SingleTypeVariantListPtr targetJointValues = getInputList("targetJointValues");

    if (jointNames->getSize()!=targetJointValues->getSize())
    {
        throw LocalException("Sizes of joint name list and joint value list do not match!");
    }

    for(int i=0; i<jointNames->getSize(); i++ )
    {
        targetJointAngles[jointNames->getElement(i)->getString()] = targetJointValues->getElement(i)->getFloat();
        ARMARX_VERBOSE << "setting joint angle for joint '" << jointNames->getElement(i)->getString() << "' to " << targetJointValues->getElement(i)->getFloat() << std::endl;
    }
    // TODO: Set Max Velocities

    // now install the condition

    //ExpressionPtr cond = Expression::create();
    Term cond;
    float tolerance = getInput<float>("targetTolerance");
    for(int i=0; i<jointNames->getSize(); i++ )
    {
        ParameterList inrangeCheck;
        inrangeCheck.push_back(new Variant(targetJointValues->getElement(i)->getFloat()-tolerance));
        inrangeCheck.push_back(new Variant(targetJointValues->getElement(i)->getFloat()+tolerance));

        //cond->add(DataFieldIdentifier(context->getKinematicUnitObserverName(),"jointangles", jointNames->getElement(i)->getString()), "inrange", inrangeCheck);
        Literal inrangeLiteral(DataFieldIdentifier(context->getKinematicUnitObserverName(),"jointangles", jointNames->getElement(i)->getString()), "inrange", inrangeCheck);
        cond = cond && inrangeLiteral;
    }

    targetReachedCondition = installCondition<EvJointTargetReached>(cond);
    ARMARX_INFO << "condition installed" << std::endl;
    timeoutEvent = setTimeoutEvent(getInput<int>("timeoutInMs"), createEvent<EvTimeout>());
    ARMARX_INFO << "timeout installed" << std::endl;
    context->kinematicUnitPrx->setJointAngles(targetJointAngles);
}



void MoveJoints::onBreak()
{
    removeCondition(targetReachedCondition);
    removeTimeoutEvent(timeoutEvent);
}



void MoveJoints::onExit()
{
    SingleTypeVariantListPtr jointNames = getInputList("jointNames");
    SingleTypeVariantListPtr targetJointValues = getInputList("targetJointValues");
    SingleTypeVariantList resultList(VariantType::Float);

    for(int i=0; i<jointNames->getSize(); i++ )
    {
        DataFieldIdentifierPtr datafield = new DataFieldIdentifier(context->getKinematicUnitObserverName(),"jointangles", jointNames->getElement(i)->getString());
        float jointAngleActual = context->kinematicUnitObserverPrx->getDataField(datafield)->getFloat();
        float jointTargetValue = targetJointValues->getElement(i)->getFloat();
        resultList.addElement(new Variant(jointAngleActual - jointTargetValue));
    }
    setOutput("jointDistancesToTargets", resultList);

    ARMARX_INFO << "output: " << StateUtilFunctions::getDictionaryString(getOutputParameters()) << std::endl;

    removeCondition(targetReachedCondition);
    removeTimeoutEvent(timeoutEvent);
}



void MoveJointsVelocityControl::defineParameters()
{
    context = getContext<RobotStatechartContext>();

    std::string configFilepathAbsolute;
    ArmarXDataPath::getAbsolutePath("HumanoidRobotAPIConfigs/stateconfigs/MotionControl.xml", configFilepathAbsolute);
    setConfigFile(configFilepathAbsolute);

    addToInputAsList("jointNames", VariantType::String, false);
    addToInputAsList("targetJointVelocities", VariantType::Float, false);

    // TODO: add when we have decided how to do it
    addToInputAsList("jointMaxSpeeds", VariantType::Float, true);
    addToInput("jointMaxSpeed", VariantType::Float, true);

    addToInput("accelerationTime", VariantType::Float, false);

    addToInput("targetJointVelocityTolerance", VariantType::Float, false);

    addToInput("timeoutInMs", VariantType::Int, false);

    addToOutputAsList("jointVelocitiesDistancesToTargets", VariantType::Float, true);
}



void MoveJointsVelocityControl::onEnter()
{
    float accelerationTime = getInput<float>("accelerationTime");
    NameValueMap targetJointVelocities;
    SingleTypeVariantListPtr jointNames = getInputList("jointNames");
    SingleTypeVariantListPtr targetVelocitiesValues = getInputList("targetJointVelocities");

    if (jointNames->getSize()!=targetVelocitiesValues->getSize())
    {
        throw LocalException("Sizes of joint name list and joint velocities list do not match!");
    }

    for(int i=0; i<jointNames->getSize(); i++)
    {
        targetJointVelocities[jointNames->getElement(i)->getString()] = targetVelocitiesValues->getElement(i)->getFloat();
        ARMARX_VERBOSE << "setting joint angle for joint '" << jointNames->getElement(i)->getString() << "' to " << targetVelocitiesValues->getElement(i)->getFloat() << std::endl;
    }
    // TODO: Set Max Velocities

    // now install the condition

    Term cond;
    float tolerance = getInput<float>("targetJointVelocityTolerance");
    for(int i=0; i<jointNames->getSize(); i++)
    {
        ParameterList inrangeCheck;
        inrangeCheck.push_back(new Variant(targetVelocitiesValues->getElement(i)->getFloat()-tolerance));
        inrangeCheck.push_back(new Variant(targetVelocitiesValues->getElement(i)->getFloat()+tolerance));

        Literal inrangeLiteral(DataFieldIdentifier(context->getKinematicUnitObserverName(),"jointvelocities", jointNames->getElement(i)->getString()), "inrange", inrangeCheck);
        cond = cond && inrangeLiteral;
    }

    targetVelocitiesReachedCondition = installCondition<EvJointVelocitiesTargetReached>(cond);
    ARMARX_INFO << "condition installed" << std::endl;
    timeoutEvent = setTimeoutEvent(getInput<int>("timeoutInMs"), createEvent<EvTimeout>());
    ARMARX_INFO << "timeout installed" << std::endl;
//    context->kinematicUnitPrx->setAccelerationTime(accelerationTime);
    context->kinematicUnitPrx->setJointVelocities(targetJointVelocities);
}



void MoveJointsVelocityControl::onExit()
{
    SingleTypeVariantListPtr jointNames = getInputList("jointNames");
    SingleTypeVariantListPtr targetJointVelocities = getInputList("targetJointVelocities");
    SingleTypeVariantList resultList(VariantType::Float);

    for(int i=0; i<jointNames->getSize(); i++)
    {
        DataFieldIdentifierPtr datafield = new DataFieldIdentifier(context->getKinematicUnitObserverName(),"jointvelocities", jointNames->getElement(i)->getString());
        float jointVelocityActual = context->kinematicUnitObserverPrx->getDataField(datafield)->getFloat();
        float jointVelocityTarget = targetJointVelocities->getElement(i)->getFloat();
        resultList.addElement(new Variant(jointVelocityActual - jointVelocityTarget));
    }
    setOutput("jointVelocitiesDistancesToTargets", resultList);

    ARMARX_INFO << "output: " << StateUtilFunctions::getDictionaryString(getOutputParameters()) << std::endl;

    removeCondition(targetVelocitiesReachedCondition);
    removeTimeoutEvent(timeoutEvent);
}



void MoveTCPPoseIK::defineParameters()
{
    context = getContext<RobotStatechartContext>();

    std::string configFilepathAbsolute;
    ArmarXDataPath::getAbsolutePath("HumanoidRobotAPIConfigs/stateconfigs/MotionControl.xml", configFilepathAbsolute);
    setConfigFile(configFilepathAbsolute);

    addToInput("KinematicChainName", VariantType::String, false);
    addToInput("targetTCPPosition", VariantType::LinkedPosition, false);
    addToInput("targetTCPOrientation", VariantType::LinkedOrientation, false);

    addToInput("tcpMaxSpeed", VariantType::Float, true);

    addToInput("targetPositionDistanceTolerance", VariantType::Float, false);
    addToInput("targetOrientationDistanceTolerance", VariantType::Float, false);
    addToInput("timeoutInMs", VariantType::Int, false);


    addToOutput("TCPDistanceToTarget", VariantType::Float, true);
    addToOutput("TCPOrientationDistanceToTarget", VariantType::Float, true);
}



void MoveTCPPoseIK::onEnter()
{
    timeoutEvent = setTimeoutEvent(getInput<int>("timeoutInMs"), createEvent<EvTimeout>());
}



void MoveTCPPoseIK::onExit()
{
    removeTimeoutEvent(timeoutEvent);
}



void MoveTCPPoseIK::defineSubstates()
{
    StatePtr moveJoints = addState<MoveJoints>("MoveJoints");
    StatePtr calculateJointAngleConfiguration = addState<CalculateJointAngleConfiguration>("CalculateJointAngleConfiguration");
    StatePtr validateJointAngleConfiguration = addState<ValidateJointAngleConfiguration>("ValidateJointAngleConfiguration");
    StatePtr movingDone = addState<SuccessState>("movingDone") ;
    StatePtr movingFailed = addState<FailureState>("movingFailed") ;
    setInitState(calculateJointAngleConfiguration, PM::createMapping()->mapFromParent("*","*"));
    addTransition<EvCalculationDone>(calculateJointAngleConfiguration, moveJoints, createMapping()->mapFromParent("*","*")->mapFromOutput("*","*"));
    addTransition<EvJointTargetReached>(moveJoints, validateJointAngleConfiguration, createMapping()->mapFromParent("*","*")->mapFromOutput("*","*"));
    addTransition<EvSuccess>(validateJointAngleConfiguration, movingDone);
    addTransitionFromAllStates<EvTimeout>(movingFailed);
    addTransitionFromAllStates<EvFailure>(movingFailed);
}



void MoveTCPTrajectory::defineParameters()
{
    context = getContext<RobotStatechartContext>();

    std::string configFilepathAbsolute;
    ArmarXDataPath::getAbsolutePath("HumanoidRobotAPIConfigs/stateconfigs/MotionControl.xml", configFilepathAbsolute);
    setConfigFile(configFilepathAbsolute);

    addToInput("KinematicChainName", VariantType::String, false);
    addToInputAsList("targetTCPPositions", VariantType::LinkedPosition, false);
    addToInputAsList("targetTCPOrientations", VariantType::LinkedOrientation, false);

    addToInput("tcpMaxSpeed", VariantType::Float, true);

    addToInput("targetPositionDistanceTolerance", VariantType::Float, false);
    addToInput("targetOrientationDistanceTolerance", VariantType::Float, false);
    addToInput("timeoutInMs", VariantType::Int, false);


    addToOutput("TCPDistanceToTarget", VariantType::Float, true);
    addToOutput("TCPOrientationDistanceToTarget", VariantType::Float, true);

    addToLocal("trajectoryPointCounterID", VariantType::ChannelRef);
}



void MoveTCPTrajectory::onEnter()
{
    timeoutEvent = setTimeoutEvent(getInput<int>("timeoutInMs"), createEvent<EvTimeout>());
    int numberOfPoints = getInputList("targetTCPPositions")->getSize();
    if (numberOfPoints == 0)
    {
        sendEvent<EvSuccess>();
    }
    else if (numberOfPoints != getInputList("targetTCPOrientations")->getSize())
    {
        throw armarx::exceptions::local::eLogicError("the number of targetTCPPositions does not equal the number of targetTCPOrientations");
        sendEvent<EvFailure>();
    }

    // create a counter to keep track of the poses that were already reached. As we dont use the event, set the max to n+1 so it will never be sent
    trajectoryPointCounterID = setCounterEvent(numberOfPoints, createEvent<EvLastPointReached>());
    setLocal("trajectoryPointCounterID", trajectoryPointCounterID.actionId.get());
}



void MoveTCPTrajectory::onExit()
{
    removeTimeoutEvent(timeoutEvent);
    removeCounterEvent(trajectoryPointCounterID);
}



void MoveTCPTrajectory::defineSubstates()
{
    StatePtr moveTCPTrajectoryNextPoint = addState<MoveTCPTrajectoryNextPoint>("MoveTCPTrajectoryNextPoint");
    StatePtr movingDone = addState<SuccessState>("movingDone");
    StatePtr movingFailed = addState<FailureState>("movingFailed");
    setInitState(moveTCPTrajectoryNextPoint, PM::createMapping()->mapFromParent("*","*"));
    addTransition<EvSuccess>(moveTCPTrajectoryNextPoint, moveTCPTrajectoryNextPoint, createMapping()->mapFromParent("*","*")->mapFromOutput("*","*"));
    addTransition<EvLastPointReached>(moveTCPTrajectoryNextPoint, movingDone, createMapping()->mapFromParent("*","*")->mapFromOutput("*","*"));
    addTransitionFromAllStates<EvTimeout>(movingFailed);
    addTransitionFromAllStates<EvFailure>(movingFailed);
}



void MoveTCPTrajectoryNextPoint::defineParameters()
{
    addToInput("KinematicChainName", VariantType::String, false);
    addToInputAsList("targetTCPPositions", VariantType::LinkedPosition, false);
    addToInputAsList("targetTCPOrientations", VariantType::LinkedOrientation, false);

    addToInput("tcpMaxSpeed", VariantType::Float, true);

    addToInput("targetPositionDistanceTolerance", VariantType::Float, false);
    addToInput("targetOrientationDistanceTolerance", VariantType::Float, false);
    addToInput("timeoutInMs", VariantType::Int, false);

    addToInput("trajectoryPointCounterID", VariantType::ChannelRef, false);


    addToLocal("targetTCPPosition", VariantType::LinkedPosition);
    addToLocal("targetTCPOrientation", VariantType::LinkedOrientation);


    addToOutput("TCPDistanceToTarget", VariantType::Float, true);
    addToOutput("TCPOrientationDistanceToTarget", VariantType::Float, true);
}



void MoveTCPTrajectoryNextPoint::onEnter()
{
    context = getContext<RobotStatechartContext>();
    ChannelRefPtr trajectoryPointCounterID = getInput<ChannelRefPtr>("trajectoryPointCounterID");
    int counterValue = trajectoryPointCounterID->getDataField("value")->getInt();

    if (counterValue < getInputList("targetTCPPositions")->getSize())
    {
        setLocal("targetTCPPosition", *VariantPtr::dynamicCast(getInputList("targetTCPPositions")->getElement(counterValue)));
        setLocal("targetTCPOrientation", *VariantPtr::dynamicCast(getInputList("targetTCPOrientations")->getElement(counterValue)));

        context->systemObserverPrx->incrementCounter(trajectoryPointCounterID);
    }
    else
    {
        sendEvent<EvLastPointReached>();
    }
}



void MoveTCPTrajectoryNextPoint::onExit()
{
}



void MoveTCPTrajectoryNextPoint::defineSubstates()
{
    StatePtr moveTCPPoseIK = addState<MoveTCPPoseIK>("MoveTCPPoseIK");
    StatePtr movingDone = addState<SuccessState>("movingDone") ;
    StatePtr movingFailed = addState<FailureState>("movingFailed") ;
    setInitState(moveTCPPoseIK, PM::createMapping()->mapFromParent("*","*"));
    addTransition<EvSuccess>(moveTCPPoseIK, movingDone);
    addTransitionFromAllStates<EvTimeout>(movingFailed);
    addTransitionFromAllStates<EvFailure>(movingFailed);
}



void CalculateJointAngleConfiguration::defineParameters()
{
    addToInput("KinematicChainName", VariantType::String, false);
    addToInput("targetTCPPosition", VariantType::LinkedPosition, false);
    addToInput("targetTCPOrientation", VariantType::LinkedOrientation, false);

    addToOutputAsList("jointNames", VariantType::String, true);
    addToOutputAsList("targetJointValues", VariantType::Float, true);
}



void CalculateJointAngleConfiguration::run()
{
    RobotStatechartContext* context = getContext<RobotStatechartContext>();
    VirtualRobot::RobotPtr robot = (RemoteRobot(context->robotStateComponent->getRobotSnapshot("CalculateTCPPoseTime"))).clone("IKCopy");

    VirtualRobot::GenericIKSolverPtr ikSolver(new VirtualRobot::GenericIKSolver(robot->getRobotNodeSet(getInput<std::string>("KinematicChainName")) ));

    VirtualRobot::LinkedCoordinate target = ArmarPose::createLinkedCoordinate(robot,
                                                                              getInput<LinkedPosition>("targetTCPPosition"),
                                                                              getInput<LinkedOrientation>("targetTCPOrientation"));

    if(!ikSolver->solve(target.getInFrame(robot->getRootNode()), VirtualRobot::IKSolver::All, 50))
    {
        sendEvent<EvFailure>();
    }
    else
    {

        SingleTypeVariantList jointNames = SingleTypeVariantList(VariantType::String);
        SingleTypeVariantList targetJointValues = SingleTypeVariantList(VariantType::Float);
        VirtualRobot::RobotNodeSetPtr kinematicChain = robot->getRobotNodeSet(getInput<std::string>("KinematicChainName") );
        for (int i = 0; i < (signed int)kinematicChain->getSize(); i++)
        {
            jointNames.addElement(new Variant(kinematicChain->getNode(i)->getName()));
            targetJointValues.addElement(new Variant(kinematicChain->getNode(i)->getJointValue()));
        }
        setOutput("targetJointValues",targetJointValues);
        setOutput("jointNames",jointNames);
        sendEvent<EvCalculationDone>();
    }
}



void ValidateJointAngleConfiguration::defineParameters()
{
    addToInput("targetTCPPosition", VariantType::LinkedPosition, false);
    addToInput("targetTCPOrientation", VariantType::LinkedOrientation, false);
    addToInput("targetPositionDistanceTolerance", VariantType::Float, false);
    addToInput("targetOrientationDistanceTolerance", VariantType::Float, false);
    addToInput("KinematicChainName", VariantType::String, false);

    addToOutput("TCPDistanceToTarget", VariantType::Float, true);
    addToOutput("TCPOrientationDistanceToTarget", VariantType::Float, true);
}



void ValidateJointAngleConfiguration::onEnter()
{
    RobotStatechartContext* context = getContext<RobotStatechartContext>();
    VirtualRobot::RobotPtr robot = (RemoteRobot(context->robotStateComponent->getRobotSnapshot("ValidateTCPPoseTime"))).clone("IKCopy");
    using namespace Eigen;
    VirtualRobot::LinkedCoordinate actualPose(robot);
    actualPose.set(getInput<std::string>("KinematicChainName"), Eigen::Vector3f(0,0,0));
    actualPose.changeFrame(robot->getRootNode());
    Vector3f actualPosition = actualPose.getPosition();
    Quaternionf actualOrientation(actualPose.getPose().block<3,3>(0,0));


    VirtualRobot::LinkedCoordinate targetPose = ArmarPose::createLinkedCoordinate(robot, getInput<LinkedPosition>("targetTCPPosition"), getInput<LinkedOrientation>("targetTCPOrientation"));
    targetPose.changeFrame(robot->getRootNode());
    Eigen::Vector3f targetPosition = targetPose.getPosition();
    Eigen::Quaternionf targetOrientation(targetPose.getPose().block<3,3>(0,0));

    float cartesianDistance = sqrtf(((targetPosition-actualPosition).dot(targetPosition-actualPosition)));
    float orientationDistance = actualOrientation.angularDistance(targetOrientation);

    setOutput("TCPDistanceToTarget", cartesianDistance);
    setOutput("TCPOrientationDistanceToTarget", orientationDistance);
    if (cartesianDistance <= getInput<float>("targetPositionDistanceTolerance") && orientationDistance <= getInput<float>("targetOrientationDistanceTolerance"))
    {
        sendEvent<EvSuccess>();
    }
    else
    {
        sendEvent<EvFailure>();
    }
}



void MotionControlTestState::defineParameters()
{
    std::string configFilepathAbsolute;
    ArmarXDataPath::getAbsolutePath("HumanoidRobotAPIConfigs/stateconfigs/MotionControl.xml", configFilepathAbsolute);
    setConfigFile(configFilepathAbsolute);

    addToInputAsList("jointNames", VariantType::String, false);
    addToInputAsList("targetJointValues", VariantType::Float, false);

    addToInput("jointMaxSpeed", VariantType::Float, true);

    addToInput("targetTolerance", VariantType::Float, false);

    addToInput("timeoutInMs", VariantType::Int, false);

    addToOutputAsList("jointDistancesToTargets", VariantType::Float, true);
}



void MotionControlTestState::defineSubstates()
{
    StatePtr moveJoints = addState<MoveJoints>("MoveJoints");
    StatePtr movingDone = addState<SuccessState>("movingDone") ;
    StatePtr movingFailed = addState<FailureState>("movingFailed") ;

    setInitState(moveJoints, PM::createMapping()->mapFromParent("*","*"));

    addTransition<EvJointTargetReached>(moveJoints, movingDone);
    addTransition<EvTimeout>(moveJoints, movingFailed);
}



void MoveHelpers::StopRobot::onEnter()
{
    RobotStatechartContext* context = getContext<RobotStatechartContext>();

    VirtualRobot::RobotPtr robot = (RemoteRobot(context->robotStateComponent->getRobotSnapshot("StopRobotTime"))).clone("StopRobotCopy");
    NameValueMap jointVelocities;

    std::vector<VirtualRobot::RobotNodePtr> nodesToCheck;
    nodesToCheck.push_back(robot->getRootNode());
    while(!nodesToCheck.empty())
    {
        VirtualRobot::RobotNodePtr current = *nodesToCheck.end();
        nodesToCheck.pop_back();
        nodesToCheck.insert(nodesToCheck.end(), current->getChildren().begin(), current->getChildren().end());
        jointVelocities[current->getName()] = 0.f;
    }

    context->kinematicUnitPrx->begin_setJointVelocities(jointVelocities);
    sendEvent<EvSuccess>();
}


